/*
 * Copyright 2007 Rasmus Toftdahl Olesen <halfdan@halfdans.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

#define G_LOG_DOMAIN "sb-junk-filter"

#include <glib.h>
#include <e-util/e-plugin.h>
#include <mail/em-junk-hook.h>
#include <camel/camel-data-wrapper.h>
#include <camel/camel-stream-fs.h>
#include <camel/camel-debug.h>
#include <camel/camel-medium.h>
#include <gconf/gconf-client.h>

#ifndef SPAMBAYES_CLIENT_BINARY
#define SPAMBAYES_CLIENT_BINARY "/usr/bin/sb_evolution_plugin_client"
#endif

#define SPAMBAYES_ERROR 3
#define PLUGIN_HEADER "X-Evolution-Spambayes-Plugin"
#define PLUGIN_HEADER_SPAM "spam"
#define PLUGIN_HEADER_HAM "ham"
#define EM_JUNK_SB_GCONF_DIR "/apps/evolution/mail/junk/spambayes"

#define d(x) (camel_debug("junk")?(x):0)
#define DEBUG_HEADER "SpamBayes Plugin: "

static const gchar em_junk_sb_gconf_dir[] = EM_JUNK_SB_GCONF_DIR;

#define EM_JUNK_SB_GCONF_DIR_LENGTH (G_N_ELEMENTS (em_junk_sb_gconf_dir) - 1)

static gint em_junk_sb_daemon_running = FALSE;
static GStaticMutex em_junk_sb_mutex = G_STATIC_MUTEX_INIT;

static void lock ()
{
    d(fprintf (stderr, DEBUG_HEADER "Locking mutex\n"));
    g_static_mutex_lock ( &em_junk_sb_mutex );
}

static void unlock ()
{    
    d(fprintf (stderr, DEBUG_HEADER "Unlocking mutex\n"));
    g_static_mutex_unlock ( &em_junk_sb_mutex );
}

static gint ping_server () {
    // Wait until server is up by ping'ing it
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-p",
        NULL
    };
    
    gint exit_code = 0;
    GError *error;
    if ( g_spawn_sync ( NULL, argv, NULL, 0, NULL, NULL, NULL, NULL, &exit_code, &error ) == TRUE && WIFEXITED (exit_code) ) {
        return WEXITSTATUS(exit_code);
    } else if ( error ) {
        printf ( DEBUG_HEADER "Could not ping server (%s)\n", error->message );
    } else {
        printf ( DEBUG_HEADER "Could not ping server (exitcode was %i)\n", exit_code );
    }
    return -1;
}

static gint tell_server_to_stop () {
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-x",
        NULL
    };
    gint exit_code = 0;
    GError *error;
    if ( g_spawn_sync ( NULL, argv, NULL, 0, NULL, NULL, NULL, NULL, &exit_code, &error ) == TRUE && WIFEXITED (exit_code) ) {
        return WEXITSTATUS(exit_code);
    } else if ( error ) {
        printf ( DEBUG_HEADER "Could not stop server (%s)\n", error->message );
    } else {
        printf ( DEBUG_HEADER "Could not stop server (exitcode was %i)\n", exit_code );
    }
    return -1;
}

static gboolean start_server ()
{
    lock ();
    if ( em_junk_sb_daemon_running == FALSE ) {
        // Wait until server is up by ping'ing it
        d(fprintf (stderr, DEBUG_HEADER "Pinging server\n"));
        gint rv = 0;
        while ( rv != 1 ) {
            rv = ping_server ();
            d(fprintf (stderr, DEBUG_HEADER " - Server ping returned %d\n", rv));
            sleep ( 0.5 );
        }
        d(fprintf (stderr, DEBUG_HEADER "Server succesfully pinged\n"));
        
        em_junk_sb_daemon_running = TRUE;
        unlock ();
        return TRUE;
    };
    unlock ();
    return FALSE;
}

static void stop_server ()
{
    lock ();
    if ( em_junk_sb_daemon_running == TRUE ) {
        tell_server_to_stop ();
        sleep ( 1 );
        em_junk_sb_daemon_running = FALSE;
    }
    unlock ();
}

static gint
pipe_to_spambayes (CamelMimeMessage *msg, gchar **argv)
{
    GPid child_pid;
    gint sb_in;
    CamelStream *stream;
    GError *err = NULL;
    gint status;
    gint waitres;
    
    if (camel_debug ("junk")) {
        int i;
        fprintf (stderr, DEBUG_HEADER "Calling client: ");
        for (i = 0; argv[i]; i++) {
            fprintf (stderr, "%s ", argv[i]);
        }
        printf ("\n");
    }
    
    d(fprintf (stderr, DEBUG_HEADER "Starting server (if necessary)\n" ));
    start_server ();
    
    d(fprintf (stderr, DEBUG_HEADER "Before spawning\n" ));
    if (!g_spawn_async_with_pipes (NULL,
                                   argv,
                                   NULL,
                                   G_SPAWN_DO_NOT_REAP_CHILD,
                                   NULL,
                                   NULL,
                                   &child_pid,
                                   &sb_in,
                                   NULL,
                                   NULL,
                                   &err)) {
        g_warning ("error occurred while spawning %s: %s",
                   argv[0],
                   err->message);
        return SPAMBAYES_ERROR;
    }
    d(fprintf (stderr, DEBUG_HEADER "Child spawned\n" ));
    
    stream = camel_stream_fs_new_with_fd (sb_in);
    camel_data_wrapper_write_to_stream (CAMEL_DATA_WRAPPER (msg), stream);
    camel_stream_flush (stream);
    camel_stream_close (stream);
    camel_object_unref (stream);
    waitres = waitpid (child_pid, &status, 0);
    if (waitres < 0 && errno == EINTR) {
        /* child process is hanging... */
        g_warning (DEBUG_HEADER "wait for sb_evolution_plugin_client child process interrupted, terminating");
        kill (child_pid, SIGTERM);
        sleep (1);
        waitres = waitpid (child_pid, &status, WNOHANG);
        if (waitres == 0) {
            /* ...still hanging, set phasers to KILL */
            g_warning (DEBUG_HEADER "sb_evolution_plugin_client child process does not respond, killing");
            kill (child_pid, SIGKILL);
            sleep (1);
            waitres = waitpid (child_pid, &status, WNOHANG);
        }
    }

    g_spawn_close_pid (child_pid);

    if (waitres >= 0 && WIFEXITED (status)) {
        return WEXITSTATUS (status);
    } else {
        return SPAMBAYES_ERROR;
    }
}

static void em_junk_sb_setting_notify (GConfClient *gconf,
                                       guint        cnxn_id,
                                       GConfEntry  *entry,
                                       void        *data)
{
    const char *key;
    GConfValue *value;

    value = gconf_entry_get_value (entry);
    if (value == NULL) {
        return;
    }

    key = gconf_entry_get_key (entry);
    g_return_if_fail (key != NULL);

    g_return_if_fail (!strncmp (key, em_junk_sb_gconf_dir, EM_JUNK_SB_GCONF_DIR_LENGTH));
    key += EM_JUNK_SB_GCONF_DIR_LENGTH;
    g_return_if_fail (*key != '/');
    ++key;

    /*
    if (strcmp (key, "port") == 0) {
        if ( em_junk_sb_port != NULL )
            g_free ( em_junk_sb_port );
        
        em_junk_sb_port = g_strdup_printf ( "%i", gconf_value_get_int (value) );
    }
    */
}

gboolean em_junk_sb_check_junk (EPlugin *ep, EMJunkHookTarget *target)
{
    CamelMimeMessage *msg = target->m;
    int rv;
    
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        NULL
    };

    d(fprintf (stderr, DEBUG_HEADER "Checking message\n"));
    
    GArray *headers = camel_medium_get_headers ( CAMEL_MEDIUM(msg) );
    const CamelMediumHeader *header;
    int i;
    for ( i = 0; i < headers->len; ++i ) {
        header = &g_array_index ( headers, CamelMediumHeader, i );
        d(fprintf (stderr, DEBUG_HEADER "Header %s: %s\n", header->name, header->value));
    }
    camel_medium_free_headers( CAMEL_MEDIUM(msg), headers );
    
    rv = pipe_to_spambayes (msg, argv);

    d(fprintf (stderr, DEBUG_HEADER "Message marked as rv = %d\n", rv));

    return (rv == 2);
}

void em_junk_sb_report_junk_untrain (CamelMimeMessage *msg)
{
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-s",
        "-u",
        NULL
    };

    d(fprintf (stderr, DEBUG_HEADER "Un-Reporting spam\n"));

    pipe_to_spambayes (msg, argv);
}

void em_junk_sb_report_non_junk_untrain (CamelMimeMessage *msg)
{
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-h",
        "-u",
        NULL
    };
    
    d(fprintf (stderr, DEBUG_HEADER "Un-Reporting ham\n"));

    pipe_to_spambayes (msg, argv);
}

void em_junk_sb_report_junk (EPlugin *ep, EMJunkHookTarget *target)
{
    CamelMimeMessage *msg = target->m;
    GArray *headers = camel_medium_get_headers ( CAMEL_MEDIUM(msg) );
    const CamelMediumHeader *_header;
    int i;
    for ( i = 0; i < headers->len; ++i ) {
        _header = &g_array_index ( headers, CamelMediumHeader, i );
        d(fprintf (stderr, DEBUG_HEADER "Header %s: %s\n", _header->name, _header->value));
    }
    camel_medium_free_headers( CAMEL_MEDIUM(msg), headers );
    
    const char *header;
    header = camel_medium_get_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER );
    if ( header ) {
        d(fprintf (stderr, DEBUG_HEADER "%s header is %s\n", PLUGIN_HEADER, header));
    }
    if ( header && strcmp(header, PLUGIN_HEADER_HAM) == 0 ) {
        em_junk_sb_report_non_junk_untrain ( msg );
    }
    
    if ( header ) {
        camel_medium_set_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER, PLUGIN_HEADER_SPAM );
    } else {
        camel_medium_add_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER, PLUGIN_HEADER_SPAM );
    }
    
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-s",
        NULL
    };

    d(fprintf (stderr, DEBUG_HEADER "Reporting spam\n"));

    pipe_to_spambayes (msg, argv);
}

void em_junk_sb_report_non_junk (EPlugin *ep, EMJunkHookTarget *target)
{
    CamelMimeMessage *msg = target->m;
    GArray *headers = camel_medium_get_headers ( CAMEL_MEDIUM(msg) );
    const CamelMediumHeader *_header;
    int i;
    for ( i = 0; i < headers->len; ++i ) {
        _header = &g_array_index ( headers, CamelMediumHeader, i );
        d(fprintf (stderr, DEBUG_HEADER "Header %s: %s\n", _header->name, _header->value));
    }
    camel_medium_free_headers( CAMEL_MEDIUM(msg), headers );

    const char *header;
    header = camel_medium_get_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER );
    if ( header ) {
        d(fprintf (stderr, DEBUG_HEADER "%s header is %s\n", PLUGIN_HEADER, header));
    }
    if ( header && strcmp(header, PLUGIN_HEADER_SPAM) == 0 ) {
        em_junk_sb_report_junk_untrain ( msg );
    }
    if ( header ) {
        camel_medium_set_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER, PLUGIN_HEADER_HAM );
    } else {
        camel_medium_add_header ( CAMEL_MEDIUM(msg), PLUGIN_HEADER, PLUGIN_HEADER_HAM );
    }
    
    gchar *argv[] = {
        SPAMBAYES_CLIENT_BINARY,
        "-h",
        NULL
    };

    d(fprintf (stderr, DEBUG_HEADER "Reporting ham\n"));

    pipe_to_spambayes (msg, argv);
}

void em_junk_sb_commit_reports (EPlugin *ep, EMJunkHookTarget *target)
{
}

int
e_plugin_lib_enable (EPluginLib *ep, int enable)
{
    GConfClient *gconf;

    if (enable != 1) {
        return 0;
    }

    lock ();

    gconf = gconf_client_get_default();

    gconf_client_add_dir (gconf,
                          em_junk_sb_gconf_dir,
                          GCONF_CLIENT_PRELOAD_ONELEVEL,
                          NULL);

    gconf_client_notify_add (gconf,
                             em_junk_sb_gconf_dir,
                             em_junk_sb_setting_notify,
                             NULL, NULL, NULL);

    //em_junk_sb_port = g_strdup_printf ( "%i", gconf_client_get_int (gconf, EM_JUNK_SB_GCONF_DIR "/port", NULL) );
    
    g_object_unref (gconf);

    unlock ();
    
    atexit ( stop_server );
    
    return 0;
}
